from __future__ import annotations
from typing import Generator
from typing import TYPE_CHECKING
from collections import deque

if TYPE_CHECKING: 
    from .schema import Column


class Row:

    def __init__(self, data: dict[str, any]) -> None:
        self._data = data
    
    def get_by(self, column: Column) -> any:
        return self._data[column.alias]


class ResultSet:

    _main: deque[dict[str, any]]
    _subs: list[deque[dict[str, any]]]

    def __init__(self) -> None:
        self._main = ...
        self._subs = []
    
    def main(self) -> Generator[Row]:
        return self._generator(self._main)
    
    def add_data(self, data: list[dict[str, any]]) -> None:
        data = deque(data)
        if self._main is ...:
            self._main = data
        else:
            self._subs.append(deque(data))
    
    def get(self, index: int) -> Generator[Row]:
        return self._generator(self._subs[index])
    
    def _generator(self, queue: deque[dict[str, any]]) -> Generator[Row]:
        while queue:
            yield Row(data = queue[0])
            queue.popleft()
    
    def is_empty(self) -> bool:
        if self._main is ... or (len(self._main) == 0):
            return True
        return False


from __future__ import annotations
from .utils import UNLOADED
from typing import Generator, TYPE_CHECKING

if TYPE_CHECKING:
    from .result import Row
    from .proxy import Proxy
    from .mapper import Model
    from .schema import Column
    from .session import Session

    
class Loader:
    _model: Model
    _session: Session

    def __init__(self, model: Model) -> None:
        self._model = model
        self._session = SessionFactory.get()
    
    def _select(self, id: any = ...) -> None:
        query = self._model.build_select()

        self._session.clear_result()

        self._session.transaction.execute_query(query)
    
        for relationship in self._model.relationships():
            if not relationship.is_lazy:
                sub_query = relationship.build_select()
                self._session.transaction.execute_query(sub_query)
    

    def _select_by_id(self, id: any) -> None:
        primary_key = self._model.identifier.column
        query = self._model.build_select().where(primary_key == id)

        self._session.clear_result()

        self._session.transaction.execute_query(query)

        for relationship in self._model.relationships():
            if not relationship.is_lazy:
                sub_query = relationship.build_select(id = id)
                self._session.transaction.execute_query(sub_query)
        
        
    def _load(self) -> list[Proxy]:
        response = []
        elements = self.load_on_pk(self._session.get_result().main(), self._model)

        for proxy in elements:
            relationships = proxy.__model__.relationships()

            for index, relationship in enumerate(relationships):
                if relationship.is_lazy:
                    setattr(proxy, relationship.attribute_name, UNLOADED)
                else:
                    associated_elements = self.load_on(
                        rows = self._session.get_result().get(index), 
                        model = relationship.reference, 
                        key = relationship.join_column.column
                    )
                    
                    try:
                        relationship.set_attribute(proxy, next(associated_elements), mirror = True)
                    except StopIteration:
                        relationship.set_null(proxy)
                    finally:
                        for associated_element in associated_elements: 
                            relationship.set_attribute(proxy, associated_element, mirror = True)
            
            response.append(proxy)

        return response

    def load(self) -> list[Proxy]:
        self._select()
        return self._load()

    def load_by_id(self, id: any) -> Proxy:
        self._select_by_id(id)
        try:
            return self._load()[0]
        except IndexError:
            raise SideralException(f"Can't find {self._model._class.__name__} register with id {id} in database.")
    
    @classmethod
    def load_on(self, rows: Generator[Row], model: Model, key: Column) -> Generator[Proxy]:
        row = ...
        key_value = ...
        try:
            row = next(rows)
            key_value = row.get_by(key)
        except StopIteration:
            return []
            
        yield self.load_element(row, model)
        for row in rows:
            if row.get_by(key) == key_value:
                yield self.load_element(row, model)
            else:
                break
    
    @classmethod
    def load_on_pk(self, rows: Generator[Row], model: Model) -> Generator[Proxy]:
        key = model.table.primary_key.column
        last_pk_value = ...

        for row in rows:
            key_value = row.get_by(key)
        
            if key_value != last_pk_value:
                yield self.load_element(row, model)
                last_pk_value = key_value
    
    @classmethod
    def load_element(self, row: Row, model: Model) -> Proxy:
        element_model = model.get_model(row)
        element = element_model.new()
        for column_prop in element_model.columns():
            setattr(
                element, 
                column_prop.attribute_name, 
                UNLOADED if column_prop.is_lazy else row.get_by(column_prop.column)
            )
        
        return element_model.proxy(element)



from .session import SessionFactory
from .errors import SideralException


from __future__ import annotations
from typing import TYPE_CHECKING, ParamSpec, TypeVar
from mysql import connector
from collections.abc import Callable
from .utils import abstract, ordered_set
from functools import wraps, partial


if TYPE_CHECKING:
    from .query import Query
    from .proxy import Proxy
    from .command import Command


class Session:
    _last_id: any
    _queue: ordered_set[Command]
    _result: ResultSet

    def __init__(self) -> None:
        self._queue = ordered_set()
        self._conn = connector.connect(option_files = 'database.conf', option_groups = ['connection'])
        self._last_id = None
        self._result = ResultSet()
    
    @property
    def transaction(self) -> Transaction:
        try:
            return self._transaction
        except AttributeError:
            self._transaction = Transaction(session = self)
            return self._transaction
    
    def __enter__(self) -> Session:
        return self
    
    def __exit__(self, *exc) -> None:
        ...
    
    def register_new(self, element: object) -> None:
        self._queue.add(
            EntityInsertCommand(
                model = Model.get(element.__class__),
                element = element
            )
        )
        
    def register_dirty(self, element: object) -> None:
        self._queue.add(
            EntityUpdateCommand(
                model = Model.get(element.__class__),
                element = element
            )
        )
    
    def register_removed(self, element: object) -> None:
        self._queue.add(
            EntityDeleteCommand(
                model = Model.get(element.__class__),
                element = element
            )
        )
    
    def load(self, _class: type, id: any = ...) -> Proxy | list[Proxy]:
        loader = Loader(model = Model(_class))
        response = loader.load() if id is ... else loader.load_by_id(id)
        return response
    
    def remove(self, element: object) -> None:
        self._queue.remove(element)

    def clear(self) -> None:
        self._queue = ordered_set()

    def flush(self) -> None:
        for command in self._queue:
            command.execute()
        
        self.clear()
    
    def queue(self, command: Command) -> None:
        self._queue.add(command)

    def get_result(self) -> ResultSet:
        return self._result
    
    def clear_result(self) -> ResultSet:
        self._last_id = None
        self._result = ResultSet()

    def get_last_id(self) -> any:
        return self._last_id

    def close(self) -> None:
        self.clear()
        self._conn.close()


class Transaction:

    def __init__(self, session: Session) -> None:
        self._session = session

    def execute_query(self, query: Query, params: dict[str, any] = {}) -> None:
        print(query)
        cursor = self._session._conn.cursor(dictionary = True)
        cursor.execute(query.statement)
        if cursor.lastrowid:
            self._session._last_id = cursor.lastrowid
        else:
            data = cursor.fetchall()
            if len(data) == 1 and (all(value in (None, ...) for value in data[0].values())):
                data = []
            
            self._session._result.add_data(data)
        
        cursor.close()

    def commit(self) -> None:
        self._result = ResultSet()
        self._session._conn.commit()


@abstract
class SessionFactory:

    _context_session: Session = None

    @classmethod
    def start(self) -> Session:
        self._context_session = Session()
        return self._context_session
    
    @classmethod
    def get(cls) -> Session:
        return cls._context_session

    @classmethod
    def restore(self, session: Session) -> Session:
        self._context_session = session


_params = ParamSpec('_params')
_return = TypeVar('_return')


# def transaction(function: Callable[_params, _return] = None, isolated: bool = False) -> Callable[_params, _return]:

#     if function is None:
#         return partial(transaction, isolated = isolated)

#     @wraps(function)
#     def isolated_transaction(*args: _params.args, **kwargs: _params.kwargs) -> _return:
#         previous_session = SessionFactory.get()
#         breakpoint()
        
#         response = ...
#         with SessionFactory.start() as session:
#             response = function(*args, **kwargs)
        
#         session.flush()
#         session.transaction.commit()
#         session.close()
#         SessionFactory.restore(previous_session)
#         return response

#     @wraps(function)
#     def joined_transaction(*args: _params.args, **kwargs: _params.kwargs) -> _return:
#         session = SessionFactory.get()
        
#         nonlocal isolated_transaction

#         if session is None:
#             return isolated_transaction(*args, **kwargs)
#         else:
#             response = ...
#             with SessionFactory.get() as session:
#                 response = function(*args, **kwargs)
#             return response

#     return isolated_transaction if isolated else joined_transaction 


# def transaction(fn = None, isolated: bool = False):

#     def wrapper(*args, **kwargs):
#         session = SessionFactory.get()

#         if session is None:
#             reponse = ...
#             with SessionFactory.start() as session:
#                 response = fn(*args, **kwargs)
#             session.flush()
#             session.transaction.commit()
#             session.close()
#             SessionFactory.restore(session)
        
#         else:
#             if isolated is True:
#                 response = SessionFactory.start()
#         ...
    
#     return wrapper


def _isolated_transaction(fn):

    def wrapper(*args, **kwargs):
        previous_session = SessionFactory.get()
        response = None
        
        with SessionFactory.start() as session:
            response = fn(*args, **kwargs)
        
        session.flush()
        session.transaction.commit()
        session.close()
        SessionFactory.restore(previous_session)
        return response
    
    return wrapper


def transaction(fn = None, isolated: bool = True):

    if fn is None:
        return partial(transaction, isolated = isolated)
    
    @wraps(fn)
    def wrapper(*args, **kwargs):
        session = SessionFactory.get()

        if isolated or session is None:
            return _isolated_transaction(fn)(*args, **kwargs)
        else:
            response = None
            with SessionFactory.get():
                response = fn(*args, **kwargs)
            return response

    return wrapper


from .result import ResultSet
from .mapper import Model
from .loader import Loader
from .command import EntityUpdateCommand, EntityDeleteCommand, EntityInsertCommand


from .annotations import entity
from .annotations import join
from .annotations import counter_join
from .annotations import join_table
from .annotations import derived_key
from .annotations import subclass
from .annotations import discriminator
from .annotations import inheritance
from .annotations import strategy
from .override import getter
from .override import setter
from .override import deleter
from .column import column
from .column import id
from .relationships import relationship
from .relationships import one_to_many
from .relationships import many_to_one
from .relationships import one_to_one
from .relationships import many_to_many
from sideral.strategy import load




from .interfaces import *
from .builder import Select
from .builder import Insert
from .builder import Update
from .builder import Delete
from .builder import Query
from .utils import Join
from .utils import format
from .condition import SQLCondition


from __future__ import annotations
from functools import partial
from enum import Enum
from sideral.utils import abstract
from collections.abc import Callable
from typing import TypeVar, ParamSpec, get_type_hints


_params = ParamSpec('_params')
_return = TypeVar('_return')


class strategy(Enum):

    SINGLE_TABLE = 'single_table'
    JOINED_TABLE = 'joined_table'



@abstract
class annotation:
    
    def __init__(self, function: Callable[_params, _return] = None, **kwargs) -> None: 
        for name, value in kwargs.items():
            setattr(self, name, value)
        
        self.__call__(function)

    def __call__(self, function: Callable[_params, _return]) -> annotation:
        self.get = function
        return self

    def __set_name__(self, owner: type, name: str) -> None:
        self.owner = owner
        self.attribute_name = name

    def __get__(self, obj, objtype) -> _return:
        if obj is None:
            return self

        return self.get(obj)    

    def __set__(self, obj, value) -> None:
        self.set(obj, value)
    
    def __delete__(self, obj) -> None:
        self.delete(obj)

    def getter(self, function: Callable[_params, _return]) -> annotation:
        self.get = function
        return self
    
    def setter(self, function: Callable[_params, _return]) -> annotation:
        self.set = function
        return self
    
    def deleter(self, function: Callable[_params, _return]) -> annotation:
        self.delete = function
        return self 
    


def join(column_name: str) -> Callable[_params, _return]:

    def wrapper(relation: relationship) -> relationship:
        relation._join_column = column_name
        relation._join_setter = True
        return relation
    
    return wrapper


def counter_join(column_name: str) -> Callable[_params, _return]:

    def wrapper(relation: relationship) -> relationship:
        relation._counter_join_column = column_name
        relation._join_setter = True
        return relation
    
    return wrapper


def join_table(name: str) -> Callable[_params, _return]:

    def wrapper(relation: relationship) -> relationship:
        relation._join_table = name
        relation._join_setter = True
        return relation
    
    return wrapper


Class = TypeVar('Class')


def entity(_class: Class = None, name: str = None) -> Class:

    if _class is None:
        return partial(entity, name = name)
    
    table = Table(name = name or _class.__name__)
    model = Model(_class)
    model.table = table

    _prepare_model_class(_class)

    return _class

        

def subclass(_class: Class) -> Class:
    SubSingleTable(Model(_class))

    _prepare_model_class(_class, sub_class = True)

    return _class


def inheritance(_class: Class = None, type: strategy = strategy.SINGLE_TABLE) -> Class:
    
    if _class is None:
        return partial(inheritance, type = type)
    
    match type:
        
        case strategy.JOINED_TABLE:
            BaseEntity(Model(_class))
        case strategy.SINGLE_TABLE:
            BaseSingleTable(Model(_class))
        case _:
            raise Exception("Unknown strategy")
    
    return _class



def derived_key(_class: Class = None, name: str = None) -> Class:

    if _class is None:
        return partial(derived_key, name = name)
    
    # _prepare_model_class(_class, sub_class = True)
    model = SubEntity(Model(_class))
    c = Column(name = name or f'id_{model.table.name}', auto_increment = False)
    model.table.add_column(c)
    model.table.primary_key = PrimaryKey(column = c)
    model.table.add_foreign_key(ForeignKey(column = c, reference = model.base.table.primary_key.column))


    return _class



def discriminator(name: str = ..., value: any = ...) -> Callable[[Class], Class]:

    def wrapper(_class: Class) -> Class:
    
        model = Model(_class)
        if name is not ...:
            model.discriminator = Column(name = name, auto_increment = False)
        if value is not ...: 
            model.discriminator_value = value

        return _class
    
    return wrapper

    




from .relationships import relationship
from .column import column
from sideral.mapper import Model
from sideral.mapper import BaseEntity
from sideral.mapper import SubEntity
from sideral.mapper import BaseSingleTable
from sideral.mapper import SubSingleTable
from sideral.schema import ForeignKey, PrimaryKey, Table
from sideral.schema import Column
from sideral.schema import ForeignKey
from .utils import _prepare_model_class


from __future__ import annotations
from sideral.utils import abstract
from typing import Generator


@abstract
class Property:

    _null_value = None

    def __init__(self, attribute_name: str, strategy: strategy.load) -> None:
        self.attribute_name = attribute_name
        self.strategy = strategy

    @property
    def is_lazy(self) -> bool:
        if self.strategy != strategy.load.EAGER: 
            return True
        return False

    def get_attribute(self, element: object) -> any:
        return getattr(element, self.attribute_name) 

    def set_attribute(self, element: Proxy, value: any) -> None:
        raise NotImplementedError
    
    def set_null(self, element: Proxy) -> None:
        setattr(element, self.attribute_name, self._null_value)

    def load(self, proxy: Proxy) -> object:
        return self.strategy(self, proxy)

    def _post_execute_action(self, result: ResultSet) -> None:
        raise NotImplementedError



class Column(Property):

    def __init__(self, attribute_name: str, strategy: strategy.load, column: schema.Column) -> None:
        super().__init__(attribute_name, strategy)
        self.column = column

    def set_attribute(self, proxy: Proxy, value: any, mirror: bool = False) -> object:
        setattr(proxy, self.attribute_name, value)
        return proxy
    
    def build_select(self, id: any = ...) -> Select:
        sql = self.column.build_select()
        
        if id is not ...:
            sql.where(self.column.table.primary_key.column == id)
        
        return sql
    
    def _post_execute_action(self, result: ResultSet) -> Generator:
        yield next(result.main()).get_by(self.column)



@abstract
class Relationship(Property):
    owner: bool = False

    def __init__(self, 
        reference: Model, 
        attribute_name: str, 
        strategy: strategy.load, 
        mapping: str, 
        table: Table, 
        join_column: ForeignKey,
    ) -> None:
        super().__init__(attribute_name, strategy)
        self.reference = reference
        self.mapping = mapping
        self.table = table
        self.join_column = join_column
    
    @property
    def relative_key_column(self) -> Column:
        raise NotImplementedError
    
    @property
    def is_many_to_many(self) -> bool:
        return False

    # def build_select(self, id: any = ...) -> Select:
    #     sql = self.reference.build_select()
    #     if not self._related:
    #         sql \
    #         .columns(self.join_column.column) \
    #         .join(self.table, type = Join.RIGHT, on = self.join_column.column == self.join_column.reference)

    #     if id is not ...:
    #         sql.where(self.join_column.column == id)
        
    #     return sql

    def _post_execute_action(self, result: ResultSet) -> Generator:
        return Loader.load_on(rows = result.main(), model = self.reference, key = self.join_column.column)



class ToOne(Relationship):

    def set_attribute(self, proxy: Proxy, value: any, mirror: bool = False) -> object:
        setattr(proxy, self.attribute_name, value)
        if mirror and self.mapping:
            value.__class__.get_property(name = self.mapping).set_attribute(value, proxy) 
        return proxy
    
    def get_key(self, element: object) -> any:
        referenced_element = getattr(element, self.attribute_name)
        if referenced_element in (None, ...):
            return None
        else:
            return self.reference.identifier.get_attribute(referenced_element)


class ToMany(Relationship):

    _null_value = []

    def __init__(
        self, 
        reference: Model, 
        attribute_name: str, 
        strategy: strategy.load, 
        mapping: str, table: Table, 
        join_column: ForeignKey,
        secondary_table: Table = None, 
        secondary_join_column: ForeignKey = None,
    ) -> None:
        super().__init__(reference, attribute_name, strategy, mapping, table, join_column)
        self.secondary_table = secondary_table
        self.secondary_join_column = secondary_join_column

    @property
    def relative_key_column(self) -> Column:
        return self.secondary_join_column.column or self.reference.table.primary_key.column
    
    @property
    def is_many_to_many(self) -> bool:
        return bool(self.secondary_table)
    
    def build_select(self, id: any = ...) -> Select:
        sql = self.reference.build_select()

        if self.secondary_table:
            sql \
            .columns([self.secondary_join_column.column, self.join_column.column]) \
            .join(self.secondary_table, type = Join.RIGHT, on = self.secondary_join_column.column == self.secondary_join_column.reference) \
            .join(self.table, type = Join.INNER, on = self.join_column.column == self.join_column.reference)

        else:
            sql \
            .columns(self.join_column.column) \
            .join(self.table, type = Join.RIGHT, on = self.join_column.column == self.join_column.reference)

        if id is not ...:
            sql.where(self.join_column.column == id) 
        
        return sql

    def set_attribute(self, proxy: Proxy, value: any, mirror: bool = False) -> object:
        try:
            getattr(proxy.__element__, self.attribute_name).append(value)
        except AttributeError:
            setattr(proxy, self.attribute_name, [value])
        finally:
            if mirror and self.mapping:
                value.__class__.get_property(name = self.mapping).set_attribute(value, proxy)
            return proxy
    
    # def _post_execute_action(self, result: ResultSet) -> Generator:
    #     return Loader.load_on(rows = result.main(), model = self.reference, key = self.join_column.column if not self.secondary_join_column else self.secondary_join_column.column)


class OneToOne(ToOne):

    def build_select(self, id: any = ...) -> Select:
        sql = self.reference.build_select() \
              .columns([self.join_column.column]) \
              .join(self.table, type = Join.RIGHT, on = self.join_column.column == self.join_column.reference)

        if id is not ...:
            sql.where(self.join_column.column == id)
        
        return sql

    @property
    def relative_key_column(self) -> Column:
        return self.reference.table.primary_key.column
    

class ManyToOne(ToOne):
    owner: bool = True

    def build_select(self, id: any = ...) -> Select:
        sql = self.reference.build_select() \
              .columns([self.join_column.column]) \
              .join(self.table, type = Join.RIGHT, on = self.join_column.reference == self.join_column.column)

        if id is not ...:
            sql.where(self.table.primary_key.column == id)
        
        return sql
    
    @property
    def relative_key_column(self) -> Column:
        return self.table.primary_key.column
    
    @property
    def is_many_to_one(self) -> bool:
        return True





from .model import Model
from sideral import schema
from sideral import strategy
from sideral.query import Select
from sideral.query import Join
from sideral.schema import Table
from sideral.schema import ForeignKey
from sideral.proxy import Proxy
from sideral.result import ResultSet
from sideral.loader import Loader


from __future__ import annotations
from sideral.utils import abstract, UNLOADED
from typing import Generator


class Model:

    __slots__ = '_class', '_table', '_columns', '_relationships', '_identifier_index', '_status'
    _instances: dict[str, Model] = {}

    def __new__(cls, _class: type) -> Model:
        try:
            return Model._instances[_class.__name__]
        except KeyError:
            model = object.__new__(cls)
            model.init(_class)
            return model
    
    def __init__(self, _class: type) -> None: ...

    def init(self, _class: type) -> None:
        self._class = _class
        self._columns = []
        self._relationships = UNLOADED
        self._identifier_index = None
        Model._instances[self.key] = self
    
    @property
    def key(self) -> str:
        return self._class.__name__

    @property
    def table(self) -> Table:
        return self._table
    
    @table.setter
    def table(self, table: Table) -> None:
        self._table = table
    
    @property
    def identifier(self) -> properties.Column:
        try:
            return self._columns[self._identifier_index]
        except KeyError:
            raise AttributeError("Id was never instatiated")
        
    
    @identifier.setter
    def identifier(self, id: properties.Column) -> None:
        self._columns.append(id)
        self._identifier_index = len(self._columns) - 1
    
    @property
    def base(self) -> Model: return None
    
    @property
    def subs(self) -> Generator[Model]: return []
    
    def columns(self, with_base: bool = True) -> Generator[properties.Column]:
        return (column for column in self._columns)
    
    def relationships(self, with_base: bool = True) -> Generator[properties.Relationship]:
        try:
            return (relationship for relationship in self._relationships)
        except UnloadedIterator:
            self._relationships = []
            for a in vars(self._class).values():
                if isinstance(a, relationships.relationship): a.setup()
            return self.relationships()

    @property
    def properties(self) -> Generator[properties.Property]:
        for prop in self.columns(): yield prop
        for prop in self.relationships(): yield prop

    @property
    def discriminator(self) -> Column:
        raise AttributeError("Model doesn't have a discriminator")
    
    @property
    def discriminator_value(self) -> any:
        raise AttributeError("Model doesn't have a discriminator value")
    
    @property
    def proxy(self) -> type[Proxy]:
        return ProxyMeta(self)
    
    @property
    def persister(self) -> Persister:
        return Persister(self)

    def new(self) -> _class:
        return object.__new__(self._class)

    def add_relationship(self, relationship: properties.Relationship) -> None:
        self._relationships.append(relationship)
    
    def add_column(self, column: properties.Column) -> None:
        self._columns.append(column)
    
    def add_sub(self, model: Model) -> None:
        raise AttributeError("Can't add sub to this model")
    
    @classmethod
    def get(cls, _class: type) -> Model:
        try:
            return cls._instances[_class.__name__]
        except KeyError:
            raise Exception(f'{_class} is not a model.')

    def get_model(self, row: Row) -> Model:
        return Model(self._class)

    def build_select(self) -> Select:
        return (
            Select()
            .columns(self.table.columns)
            .from_(self.table)
            .defer(column_prop.column for column_prop in self.columns() if column_prop.is_lazy)
        )
    
    def build_insert(self) -> Insert:
        return Insert().into(self._table).on_duplicate_key()

    def build_update(self) -> Update:
        return Update().table(self._table)
    
    def build_delete(self) -> Delete:
        return Delete().from_(self._table)

@abstract
class ModelDecorator(Model):
    
    __slots__ = '_model'

    def __new__(cls, model: Model) -> Model:
        new_model = object.__new__(cls)
        new_model.init(model)
        return new_model

    def init(self, model: Model) -> None:
        self._model = model
        self._class = model._class
        Model._instances[model.key] = self

    @property
    def key(self) -> str:
        return self._model.key

    @property
    def table(self) -> Table:
        return self._model.table
    
    @table.setter
    def table(self, table: Table) -> None:
        self._model.table = table    
    
    @property
    def identifier(self) -> properties.Column:
        return self._model.identifier
    
    @identifier.setter
    def identifier(self, id: properties.Column) -> properties.Column:
        self._model.identifier = id

    def columns(self, with_base: bool = True) -> Generator[properties.Column]:
        return self._model.columns(with_base = with_base)
    
    def relationships(self, with_base: bool = True) -> Generator[properties.Relationship]:
        return self._model.relationships(with_base = with_base)

    @property
    def properties(self) -> Generator[properties.Property]:
        return self._model.properties

    @property
    def base(self) -> Model:
        return self._model.base
    
    @property
    def subs(self) -> Generator[Model]:
        return self._model.subs
    
    @property
    def discriminator(self) -> Column:
        return self._model.discriminator
    
    @property
    def discriminator_value(self) -> any:
        return self._model.discriminator_value
    
    @property
    def persister(self) -> Persister:
        return self._model.persister
    
    def add_relationship(self, relationship: properties.Relationship) -> None:
        return self._model.add_relationship(relationship)

    def add_column(self, column: properties.Column) -> None:
        return self._model.add_column(column)

    def add_sub(self, model: Model) -> None:
        return self._model.add_sub(model)

    def get_model(self, row: Row) -> Model:
        return self._model.get_model(row)

    def build_select(self) -> Select:
        return self._model.build_select()
    
    def build_insert(self) -> Insert:
        return self._model.build_insert()

    def build_update(self) -> Update:
        return self._model.build_update()

    def build_delete(self) -> Delete:
        return self._model.build_delete()
    
    def new(self) -> object:
        return self._model.new()


class BaseEntity(ModelDecorator):

    __slots__ = '_subs'
    
    def init(self, model: Model) -> None:
        super().init(model)
        self._subs = []

    @property
    def subs(self) -> Generator[Model]:
        for _class in self._subs:
            model = Model(_class)
            yield model
            for _sub in model.subs: yield _sub
    
    def add_sub(self, model: Model) -> None:
        self._subs.append(model._class)

    def build_select(self) -> Select:
        sql = super().build_select()
        for model in self.subs:
            table = model.table
            foreign_pk = table.get_foreign_key(name = table.primary_key.column.name)

            sql.columns(table.columns).defer(column_prop.column for column_prop in model.columns() if column_prop.is_lazy)
            sql.join(table, type = Join.LEFT, on = foreign_pk.column == foreign_pk.reference)

        return sql

    def get_model(self, row: Row) -> Model:
        model = super().get_model(row)
        for _class in self._subs:
            sub_model = Model(_class)
            if row.get_by(sub_model.table.primary_key.column):
                return sub_model.get_model(row)
        
        return model


class SubEntity(ModelDecorator):

    __slots__ = '_base'

    def init(self, model: Model) -> None:
        super().init(model)
        self.base.add_sub(self)
    
    @property
    def identifier(self) -> properties.Column:
        return self.base.identifier

    def columns(self, with_base: bool = True) -> Generator[properties.Column]:
        if with_base:
            for column in self.base.columns(): yield column
        
        for column in super().columns(): yield column

    def relationships(self, with_base: bool = True) -> list[properties.Relationship]:
        if with_base:
            for relationship in self.base.relationships(): yield relationship 
        
        for relationship in super().relationships(): yield relationship

    @property
    def properties(self) -> Generator[properties.Property]:
        for column in self.columns(): yield column
        for relationship in self.relationships(): yield relationship

    @property
    def base(self) -> Model:
        return Model(self._class.__bases__[0])

    @property
    def bases(self) -> Generator[Model]:
        base = self.base
        while base:
            yield base
            base = base.base
        
    @property
    def persister(self) -> Persister:
        return JoinedTablePersister(self)        

    def build_select(self) -> Select:
        sql = super().build_select()
        foreign_pk = self.table.get_foreign_key(name = self.table.primary_key.column.name)
        for model in self.bases:
            table = model.table

            sql.columns(table.columns).defer(column_prop.column for column_prop in model.columns() if column_prop.is_lazy)
            sql.join(table, type = Join.INNER, on = foreign_pk.column == foreign_pk.reference)
            
            foreign_pk = table.get_foreign_key(name = table.primary_key.column.name)

        return sql


class BaseSingleTable(ModelDecorator):
    
    __slots__  = '_subs', '_discriminator', '_discriminator_value'

    def init(self, model: Model) -> None:
        super().init(model)
        self._discriminator = None
        self._discriminator_value = None
        self._subs = {}

    @property
    def discriminator(self) -> Column:
        return self._discriminator
    
    @discriminator.setter
    def discriminator(self, column: Column) -> None:
        self.table.add_column(column)
        self._discriminator = column

    @property
    def discriminator_value(self) -> any:
        return self._discriminator_value

    @discriminator_value.setter
    def discriminator_value(self, value: any) -> None:
        self._discriminator_value = value
    
    @property
    def persister(self) -> Persister:
        return SingleTablePersister(self)

    def add_sub(self, model: Model) -> None:
        self._subs[model.discriminator_value] = model._class

    def get_model(self, row: Row) -> Model:
        value = row.get_by(self._discriminator)
        if self._discriminator_value == value:
            return self
        try:
            return Model(self._subs[value])
        except KeyError:
            raise Exception("Discriminator value doesn't have a equivalent class")


class SubSingleTable(ModelDecorator):

    __slots__ = '_discriminator_value'

    def init(self, model: Model) -> None:
        super().init(model)
        self._discriminator_value = None
    
    @property
    def identifier(self) -> properties.Column:
        return self.base.identifier

    @property
    def base(self) -> Model:
        return Model(super()._class.__bases__[0])
    
    @property
    def properties(self) -> Generator[properties.Property]:
        for column in self.columns(): yield column
        for relationship in self.relationships(): yield relationship

    @property
    def discriminator_value(self) -> any:
        return self._discriminator_value
    
    @discriminator_value.setter
    def discriminator_value(self, value: any) -> None:
        self._discriminator_value = value
        self.base.add_sub(self)
    
    @property
    def persister(self) -> Persister:
        return SingleTablePersister(self)

    @property
    def table(self) -> Table:
        return self.base.table

    @table.setter
    def table(self, table: Table) -> None:
        raise AttributeError("Can't change this class' table")
    
    def columns(self, with_base: bool = True) -> Generator[properties.Column]:
        if with_base:
            for column in self.base.columns(): yield column
        
        for column in super().columns(): yield column

    def build_select(self) -> Select:
        return self.base.build_select().where(self.base.discriminator == self._discriminator_value)
    
    def build_insert(self) -> Insert:
        sql = self.base.build_insert().values([(self.base.discriminator, self._discriminator_value)])
        return sql
    
    def build_delete(self) -> Delete:
        return Delete().from_(self.table)


from . import properties
from sideral.annotation import relationships
from sideral.schema import Table
from sideral.schema import Column
from sideral.result import Row
from sideral.query import Select, Join, Insert, Update, Delete
from sideral.proxy import Proxy, ProxyMeta
from sideral.errors import UnloadedIterator
from sideral.persister import Persister, JoinedTablePersister, SingleTablePersister


from __future__ import annotations
from collections.abc import Callable
from typing import ParamSpec, TypeVar


_params = ParamSpec('_params')
_return = TypeVar('_return')


class _Override:

    _functions: dict[str, list[Callable[_params, _return]]] = {}

    def __init__(self, name: str,) -> None:
        self._name = name
        self._return = int

    def add(self, function: Callable[_params, _return], index: int) -> _Override:
        if self._name in self._functions:
            self._functions[self._name][index] = function
        else:
            self._functions[self._name] = [None, None, None]
            self._functions[self._name][index] = function

        return self

    def __set_name__(self, owner, name) -> property:
        setattr(owner, name, property(*_Override._functions[name]))
        del self._functions[name]



def getter(function: Callable[_params, _return]):
    return _Override(function.__name__).add(function, index = 0)

def setter(function: Callable[_params, _return]):
    return _Override(function.__name__).add(function, index = 1)
    
def deleter(function: Callable[_params, _return]):
    return _Override(function.__name__).add(function, index = 2)



from __future__ import annotations
from .utils import interface

@interface
class Command:
    
    def execute(self) -> None:
        ...

class EntityPersistCommand(Command):
    _model: Model
    _element: _model._class

    def __init__(self, model: Model, element: _model._class) -> None:
        self._model = model
        self._element = element
    
    def __hash__(self) -> int:
        return hash(self._element)


class EntityDeleteCommand(EntityPersistCommand):
    
    def execute(self) -> None:
        self._model.persister.delete(self._element)
        DeleteEntityEvent(self._model).post_event(element = self._element)


class EntityUpdateCommand(EntityPersistCommand):
    
    def execute(self) -> None:
        self._model.persister.update(self._element)
        UpdateEntityEvent(self._model).post_event(element = self._element)


class EntityInsertCommand(EntityPersistCommand):
    
    def execute(self) -> None:
        self._model.persister.insert(self._element)
        InsertEntityEvent(self._model).post_event(element = self._element)


class AssociationPersistCommand(Command):
    _model: Model
    _relationship: Relationship
    _element: _model._class
    
    def __init__(self, model: Model, relationship: Relationship, element: _model._class) -> None:
        self._model = model
        self._relationship = relationship
        self._element = element

    def execute(self) -> None:
        persister = AssociationPersister(
            relationship = self._relationship,
            model = self._model
        )
        association_list = self._relationship.get_attribute(self._element)
        key = self._model.identifier.get_attribute(self._element)
        persister.persist(association_list, key)


from .mapper import Model
from .mapper import Relationship
from .persister import AssociationPersister
from .event import InsertEntityEvent, UpdateEntityEvent, DeleteEntityEvent


from __future__ import annotations
from typing import Iterator
from .utils import Join


class Query:
    
    sub_queries: list[Query]

    def __str__(self) -> str:
        return self.statement

    @property
    def statement(self) -> str: raise NotImplementedError

    def sub_query(self, query: Query) -> Query: raise NotImplementedError


class Select(Query):

    def __init__(self) -> None:
        self._join = ''
        self._where = []
        self._tables = {}
        self._from_clause = ''
        self._order_by = ''
        self._columns = {}
        self.sub_queries = []
        self._index = 0
    
    def columns(self, columns: Column | Iterator[Column]) -> Select:
        columns = (columns,) if isinstance(columns, Column) else columns
        
        for column in columns:
            try:
                self._columns[column].append(column)
            except KeyError:
                self._columns[column] = [column]

        return self

    def from_(self, table: Table) -> Select:
        self._from_clause += f' from {table} as {self._add_table(table)}'
        return self

    def join(self, table: Table, on: SQLCondition, type: Join = Join.INNER) -> Select:
        self._from_clause += f' {type.value} {table} as {self._add_table(table)} on {on.to_string(self, on = True)}'
        return self
    
    def where(self, *conditions: list[str]) -> Select:
        for condition in conditions: 
            self._where.append(condition)
        return self

    def order_by(self, column: Column) -> Select:
        self._index = -1
        self._order_by = f' order by {column.get_alias(self)}'
        self._index = 0
        return self

    def defer(self, columns: Iterator[Column]) -> Select:
        for column in columns: 
            del self._columns[column]
        return self
    
    def sub_query(self, query: Query) -> Query:
        self.sub_queries.append(query)

    def _add_table(self, table: Table) -> str:
        table_alias = ...
        if table in self._tables:
            aliases = self._tables[table]
            table_alias = table.get_alias(len(aliases) + 1)
            aliases.append(table_alias)
        else:
            table_alias = table.get_alias(1)
            self._tables[table] = [table_alias]

        return table_alias

    @property
    def statement(self) -> str:
        sql = 'select' 

        for columns in self._columns.values():
            for index, column in enumerate(columns):
                self._index = index
                sql += f' {column.get_alias(self)} as {column.alias},'
        
        self._index = -1
        sql = sql[:-1]
        sql += self._from_clause
        if self._where:
            sql += ' where'
            sql += ''.join(f' {condition.to_string(self)} and' for condition in self._where)[:-4]

        sql += self._order_by
        sql +=  '; '
        for query in self.sub_queries: sql += f'{query};'
        self._index = 0
        return sql 


class Insert(Query):
    
    def __init__(self) -> None:
        self._values = []
        self._on_duplicate = False

    def into(self, table: Table) -> Insert:
        self._table = table
        return self
    
    def values(self, values: Iterator[tuple[Column, any]]) -> Insert:
        for column, value in values:
            self._values.append((column, value))
        return self

    def on_duplicate_key(self) -> Insert:
        self._on_duplicate = True
        return self
    
    @property
    def statement(self) -> str:
        sql = 'insert'
        sql += f' into {self._table} ('
        sql += ''.join(f'{value[0]}, ' for value in self._values)[:-2]
        sql += ') values ('
        sql += ''.join(f'{format(value[1]) if value[1] not in (None, ...) else "default"}, ' for value in self._values)[:-2]
        sql += ')'

        if self._on_duplicate:
            sql += ' on duplicate key update '
            sql += ''.join(f'{value[0] == value[1]}, ' for value in self._values)[:-2]
        
        sql += '; '
        return sql

class Update(Query):
    
    def __init__(self) -> None:
        self._where = []

    def table(self, table: Table) -> Update:
        self._table = table
        return self

    def set(self, columns: list[tuple[Column, any]]) -> Update:
        self._columns = columns
        return self
    
    def where(self, *conditions: list[str]) -> Update:
        for condition in conditions:
            self._where.append(condition)
        return self
    
    @property
    def statement(self) -> str:
        sql = f'update {self._table} set '
        sql += ''.join(f'{column[0] == column[1]}, ' for column in self._columns)[:-2]

        if self._where:
            sql += ' where'
            sql += ''.join(f' {condition} and' for condition in self._where)[:-4]
        sql += '; '
        return sql


class Delete(Query):
    
    def __init__(self) -> None:
        self._where = []

    def from_(self, table: Table) -> Delete:
        self._from = table
        return self
    
    def where(self, *conditions: list[str]) -> Delete:
        for condition in conditions:
            self._where.append(condition)
        return self

    @property
    def statement(self) -> str:
        sql = 'delete '
        sql += f'from {self._from}'

        if self._where:
            sql += ' where'
            sql += ''.join(f' {condition} and' for condition in self._where)[:-4]
        sql += '; '
        return sql


from sideral.schema import Table, Column
from .utils import format
from .condition import SQLCondition


from .events import Event
from .events import EntityEvent
from .events import InsertEntityEvent
from .events import UpdateEntityEvent
from .events import DeleteEntityEvent
from .listeners import PersistAssociationEventListener


from __future__ import annotations
from typing import TYPE_CHECKING
from enum import Enum
from .utils import UNLOADED
from .session import transaction, SessionFactory

if TYPE_CHECKING:
    from .proxy import Proxy
    from .mapper import Property


def eager(prop: Property, proxy: Proxy) -> any:
    return getattr(proxy.__element__, prop.attribute_name)


@transaction
def lazy(prop: Property, proxy: Proxy) -> any:
    attribute_value = prop.get_attribute(proxy.__element__)

    if attribute_value is UNLOADED:
        model = proxy.__model__
        sql = prop.build_select(id = getattr(proxy, model.identifier.attribute_name))

        session = SessionFactory.get()
        session.clear_result()
        session.transaction.execute_query(sql)
        result = session.get_result()

        if result.is_empty():
            prop.set_null(proxy)
        else:
            for value in prop._post_execute_action(result): 
                prop.set_attribute(proxy, value, mirror = True)
        
        return getattr(proxy.__element__, prop.attribute_name)
    else:
        return attribute_value


def no(prop: Property, proxy: Proxy) -> any: 
    attribute_value = getattr(proxy.__element__, prop.attribute_name)
    if attribute_value is UNLOADED:
        return prop._null_value
    else:
        return attribute_value


class  load(Enum):

    EAGER = eager
    LAZY = lazy
    NO = no





from collections.abc import Callable
from typing import ParamSpec, TypeVar

Class = TypeVar('Class')
_params = ParamSpec('_params')
_return = TypeVar('_return')

class private_list:
    ...


def set_persister_slot(_class: type):
    if '__slots__' in vars(_class):
        _class.__slots__.append('__persister__')


def constructor(constructor: Callable[_params, _return]) -> Callable[_params, _return]:
    
    def wrapper(*args: _params.args, **kwargs: _params.kwargs) -> _return:
        args[0].__persister__ = 'new persister'
        return constructor(args, kwargs)

    return wrapper

def getter():
    ...

def setter():
    ...

def deleter():
    ...

def _hash(attribute_name: str) -> int:
    
    def wrapper(self) -> int:
        return hash(getattr(self, attribute_name))
    
    return wrapper

def _eq(self, other: any) -> bool:
    _class = self.__class__
    
    if isinstance(other, _class):
        for column in _class.__columns__:
            try:
                if getattr(self, column) != getattr(other, column):
                    return False
            except AttributeError:
                return False
    else:
        return False

    return True


def _prepare_model_class(_class: type, sub_class: bool = False) -> None:
    _class.__eq__ = _eq
    
    for column in _class.__columns__:
        getattr(_class, column).map()
    
    if sub_class:
        _class.__columns__.extend(_class.__bases__[0].__columns__)
    


from __future__ import annotations
from sideral.utils import interface, abstract


@interface
class EventListener:
    
    def __call__(self, *args, **kwds) -> None:
        raise NotImplementedError


class PersistAssociationEventListener(EventListener):

    def __init__(self, relationship: Relationship) -> None:
        self._relationship = relationship

    def __call__(self, model: Model, element: object) -> None:
        command = AssociationPersistCommand(model = model, relationship = self._relationship, element = element)
        SessionFactory.get().queue(command)


from sideral.mapper import Model
from sideral.mapper import Relationship
from sideral.command import AssociationPersistCommand
from sideral.session import SessionFactory



    
    


class private_collection:

    def __init__(self, annotation, l: list) -> None:
        self._annotation = annotation
        self._l = l

    def __getattribute__(self, name: str) -> any:
        return getattr(self._l, name)
    
    def __setitem__(self, index, item) -> None:
        # self._annotation.update_row()
        self._l[index] = item

    def __delitem__(self, index) -> None:
        # self._annotation.delete_row()
        del self._l[index]

    def append(self, item) -> None:
        # self._annotation.insert_row()
        return self._l.append(item)
    
    def insert(self, position, item) -> None:
        # self._annotation.insert_row()
        return self._l.extend(position, item)
    
    def extend(self, l) -> None:
        # self._annotation.insert_rows()
        return self._l.extend(l)


class private_list:

    def __init__(self, annotation, l: list) -> None:
        self._annotation = annotation
        self._l = l


from __future__ import annotations
from .utils import metaclass
from typing import NewType, TYPE_CHECKING

if TYPE_CHECKING:
    from .mapper import Model    
    from .mapper import Property


def _constructor(self, element: object) -> None:
    object.__setattr__(self, '__element__', element)

def _getter(self, item):
    try:
        return self.__properties__[item].load(self)
    except KeyError: 
        raise AttributeError

def _setter(self, item, value):
    setattr(self.__element__, item, value)

def _str(self) -> str:
    return str(self.__element__)

def _eq(self, other) -> bool:
    return self.__element__.__eq__(other)

def _eq(self, other) -> bool:
    _class = self.__class__.__bases__[0]
    if isinstance(other, self.__class__.__bases__[0]):
        for column in _class.__columns__:
            if getattr(self, column) != getattr(other, column):
                return False
    else:
        return False
    
    return True


@metaclass
class ProxyMeta(type):

    __model__: Model
    __element__: __model__._class
    __properties__: dict[str, Property]

    _instances: dict[str, Proxy] = {}

    def __new__(cls, model: Model) -> None:
        try:
            return cls._instances[model._class.__name__]
        except KeyError:

            attrs = {
                '__model__': model,
                '__properties__' : {prop.attribute_name: prop for prop in model.properties},
                '__init__': _constructor,
                '__getattr__': _getter,
                '__setattr__': _setter,
                '__str__': lambda self: str(self.__element__),
                '__eq__': _eq,
                '__hash__': lambda self: self.__element__.__hash__()
            }            

            return super().__new__(cls, f'{model._class.__name__}', (model._class,), attrs)
    
    def get_property(cls, name: str) -> Property:
        return cls.__properties__[name]



Proxy = NewType("Proxy", ProxyMeta)


from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .mapper import Model, Relationship



class Persister:
    _session: Session
    _model: Model

    def __init__(self, model: Model) -> None:
        self._model = model
        self._session = SessionFactory.get()
    
    def build_insert(self, element: object) -> Insert:
        return (
            self._model.build_insert()
            .values((column.column, column.get_attribute(element)) for column in self._model.columns(with_base = False))
            .values((relationship.join_column.column, relationship.get_key(element))
                     for relationship in self._model.relationships(with_base = False) 
                     if relationship.owner
            )
        )
    
    def build_update(self, element: object) -> Update:
        primary_key = self._model.table.primary_key.column
        return (
            self._model.build_update()
            .set((column.column, column.get_attribute(element)) for column in self._model.columns(with_base = False))
            .set((relationship.join_column.column, relationship.get_key(element)) 
                  for relationship in self._model.relationships(with_base = False) 
                  if relationship.owner
            )
            .where(primary_key == self._model.identifier.get_attribute(element))
        )
    
    def build_delete(self, element: object) -> Delete:
        primary_key = self._model.table.primary_key.column
        return (
            self._model.build_delete()
            .where(primary_key == self._model.identifier.get_attribute(element))
        )
    
    def insert(self, element: object) -> None:
        self._session.transaction.execute_query(self.build_insert(element))
        id_value = self._session.get_last_id()

        if id_value is not None:
            setattr(element, self._model.identifier.attribute_name, id_value)

        return id_value
    
    def update(self, element: object) -> None:
        self._session.transaction.execute_query(self.build_update(element))
    
    def delete(self, element: object) -> None:
        self._session.transaction.execute_query(self.build_delete(element))



class JoinedTablePersister(Persister):

    def build_insert(self, element: object) -> Insert:
        primary_key = self._model.table.primary_key.column
        return super().build_insert(element).values([(primary_key, self._model.identifier.get_attribute(element))])

    def insert(self, element: object) -> None:
        self._model.base.persister.insert(element)
        super().insert(element)
    
    def update(self, element: object) -> None:
        self._model.base.persister.update(element)
        return super().update(element)

    def delete(self, element: object) -> None:
        super().delete(element)
        self._model.base.persister.delete(element)



class SingleTablePersister(Persister):
    
    def build_insert(self, element: object) -> Insert:
        return self._model.build_insert() \
                          .values((column.column, column.get_attribute(element)) for column in self._model.columns())


class AssociationPersister:

    def __init__(self, relationship: Relationship, model: Model) -> None:
        self._model = model
        self._relationship = relationship
        self._session = SessionFactory.get()
    
    def build_insert(self, key: any) -> Insert:
        join_column = self._relationship.join_column.column
        return Insert() \
                .into(join_column.table) \
                .values([
                   (join_column, key) 
                ]) \
                .on_duplicate_key()

    def build_update(self, key: any) -> Update:
        join_column = self._relationship.join_column.column
        return Update() \
                .table(join_column.table) \
                .set([
                    (join_column, key)
                ])

    def build_delete(self, key: any) -> Delete:
        join_column = self._relationship.join_column.column
        return Delete() \
                .from_(join_column.table) \
                .where(join_column == key)

    def save(self, element: object, key: any) -> None:
        sql = ...
        pk = self._relationship.reference.identifier.get_attribute(element)

        if self._relationship.is_many_to_many:
            column = self._relationship.secondary_join_column.column
            sql = self.build_insert(key) \
                    .values([(column, pk)])
        else:
            column = self._relationship.reference.table.primary_key.column
            sql = self.build_update(key) \
                    .where(column == pk)

        self._session.transaction.execute_query(sql)

    def delete(self, element: object, key: any) -> None:
        sql = ...
        pk = self._relationship.reference.identifier.get_attribute(element)
        
        if self._relationship.is_many_to_many:
            column = self._relationship.secondary_join_column.column
            sql = self.build_delete(key) \
                    .where(column == pk)
        else:
            column = self._relationship.reference.table.primary_key.column
            sql = self.build_update(key) \
                    .where(column == pk)
        
        self._session.transaction.execute_query(sql)
    
    def reset(self, key: any) -> None:
        sql = self.build_delete(key) \
                if self._relationship.is_many_to_many \
                else self.build_update(None).where(self._relationship.join_column.column == key)

        self._session.transaction.execute_query(sql)

    def recreate(self, elements: list[object], key: any) -> None:
        elements = set(elements)
        persisted_elements = self.load_relationship(key)
        
        for element in persisted_elements.difference(elements):
            self.delete(element, key)
        
        for element in elements.difference(persisted_elements):
            self.save(element, key)
    
    def persist(self, elements: list[object], key: any) -> None:
        if elements in (None, ..., []):
            self.reset(key)
        elif not hasattr(elements, '__iter__'):
            self.recreate([elements], key)
        else:
            self.recreate(elements, key)
    
    def load_relationship(self, id: any) -> set:
        self._session.clear_result()
        self._session.transaction.execute_query(self._relationship.build_select(id = id))
        
        return set(self._relationship._post_execute_action(self._session.get_result()))     




from .query import Insert, Update, Delete
from .session import Session, SessionFactory


from __future__ import annotations
from collections.abc import Callable
from typing import ParamSpec, TypeVar, get_args, get_type_hints

from .annotations import annotation
from sideral.utils import abstract
from sideral.strategy import  load


_params = ParamSpec('_params')
_return = TypeVar('_return')


@abstract
class relationship(annotation):

    def __init__(self, function: Callable[_params, _return] = None, mapping: str = None, master: bool = False, load: load =  load.LAZY) -> None:
        super().__init__(function, mapping = mapping, master = master, _join_column = None, load = load)

    @property
    def reference(self) -> type:
        return get_type_hints(self.get)['return']
    
    def setup(self) -> None:
        ...



class one_to_many(relationship):

    @property
    def reference(self) -> type:
        try: 
            return get_args(super().reference)[0]
        except NameError:
            raise ReferenceNotInstantiated
    
    @property
    def join_column(self) -> str:
        default_join_column = f'id_{Model(self.owner).table.name}'
        try:
            return self._join_column or getattr(self.reference, self.mapping)._join_column or default_join_column
        except TypeError:
            return default_join_column


    def setup(self) -> None:
        model = Model(self.owner)
        reference = Model(self.reference)
        
        column = Column(name = self.join_column, auto_increment = False)
        fk = ForeignKey(column = column, reference = model.table.primary_key.column)
        
        reference.table.add_column(column)
        reference.table.add_foreign_key(fk)
        
        relationship = properties.ToMany(
            reference = reference, 
            attribute_name = self.attribute_name, 
            strategy = self.load,
            mapping = self.mapping,
            table = model.table,
            join_column = fk,
        )
        model.add_relationship(relationship)
        
        if self.master:
            InsertEntityEvent(model).add_listener(PersistAssociationEventListener(relationship))



class many_to_one(relationship):

    @property
    def reference(self) -> type:
        try: 
            return super().reference
        except NameError: 
            raise ReferenceNotInstantiated

    @property
    def join_column(self) -> str:
        default_join_column = f'id_{Model(self.reference).table.name}'
        try:
            return self._join_column or getattr(self.reference, self.mapping)._join_column or default_join_column
        except TypeError:
            return default_join_column
    

    def setup(self) -> None:
        model = Model(self.owner)
        reference = Model(self.reference)

        column = Column(name = self.join_column, auto_increment = False)
        fk = ForeignKey(column = column, reference = reference.table.primary_key.column)

        model.table.add_column(column)
        model.table.add_foreign_key(fk)

        relationship = properties.ManyToOne(
            reference = reference,
            attribute_name = self.attribute_name,
            strategy = self.load,
            mapping = self.mapping,
            table = model.table,
            join_column = fk,
        )

        if not self.master:
            relationship.owner = False

        model.add_relationship(relationship)



class one_to_one(relationship):
 
    def __init__(self, function: Callable[_params, _return] = None, mapping: str = None, master: bool = False, load: load =  load.LAZY, owner: bool = False) -> None:
        super().__init__(function, mapping, master, load)
        self._join_owner = owner

    # def __set_name__(self, owner: type, name: str) -> None:
    #     super().__set_name__(owner, name)
    #     try:
    #         self._join_owner = not getattr(self.reference, self.mapping)
    #     except (ReferenceNotInstantiated, TypeError):
    #         self._join_owner = self._join_owner

    @property
    def reference(self) -> type:
        try: 
            return super().reference
        except NameError: 
            raise ReferenceNotInstantiated

    @property
    def join_column(self) -> str:
        default_join_column = f'id_{Model(self.owner).table.name if self.master else Model(self.reference).table.name}'
        try:
            return self._join_column or getattr(self.reference, self.mapping)._join_column or default_join_column
        except TypeError:
            return default_join_column
        
    def setup(self) -> None:
        model = Model(self.owner)
        reference = Model(self.reference)

        column = Column(name = self.join_column, auto_increment = False)

        relationship = ...

        if self._join_owner:
            model.table.add_column(column)
            fk = ForeignKey(column = column, reference = reference.table.primary_key.column)
            relationship = properties.ManyToOne(
                reference = reference,
                attribute_name = self.attribute_name,
                strategy = self.load,
                mapping = self.mapping,
                table = model.table,
                join_column = fk,
            )
        else:
            reference.table.add_column(column)
            fk = ForeignKey(column = column, reference = model.table.primary_key.column)
            relationship = properties.OneToOne(
                reference = reference,
                attribute_name = self.attribute_name,
                strategy = self.load,
                mapping = self.mapping,
                table = model.table,
                join_column = fk,
            )

            if self.master:
                InsertEntityEvent(model).add_listener(PersistAssociationEventListener(relationship))

        model.add_relationship(relationship)




class many_to_many(relationship):

    def __init__(self, function: Callable[_params, _return] = None, mapping: str = None, master: bool = False, load:  load =  load.LAZY) -> None:
        super().__init__(function, mapping = mapping, master = master, load = load)
        self._counter_join_column = None
        self._join_table = None

    @property
    def reference(self) -> type:
        try: 
            return get_args(super().reference)[0]
        except NameError: 
            raise ReferenceNotInstantiated

    @property
    def join_table(self) -> str:
        table_names = [Model(self.owner).table.name, Model(self.reference).table.name]
        table_names.sort()
        default_join_table = f'{table_names[0]}_{table_names[1]}'

        try:
            return self._join_table or getattr(self.reference, self.mapping)._join_table or default_join_table
        except TypeError:
            return default_join_table
        
    @property
    def join_column(self) -> str:
        default_join_column = f'id_{Model(self.owner).table.name}'
        try:
            return self._join_column or getattr(self.reference, self.mapping)._counter_join_column or default_join_column
        except TypeError:
            return default_join_column
    
    @property
    def counter_join_column(self) -> str:
        default_counter_join_column = f'id_{Model(self.reference).table.name}'
        try:
            return self._counter_join_column or getattr(self.reference, self.mapping)._join_column or default_counter_join_column
        except TypeError:
            return default_counter_join_column

    def setup(self) -> None:
        model = Model(self.owner)
        reference = Model(self.reference)

        join_table = Table(name = self.join_table)
        column = Column(name = self.join_column, auto_increment = False)
        counter_column = Column(name = self.counter_join_column, auto_increment = False)
        
        join_table.add_column(column)
        join_table.add_column(counter_column)

        fk = ForeignKey(column = column, reference = model.table.primary_key.column)
        counter_fk = ForeignKey(column = counter_column, reference = reference.table.primary_key.column)

        join_table.add_foreign_key(fk)
        join_table.add_foreign_key(counter_fk)

        relationship = properties.ToMany(
            reference = reference,
            attribute_name = self.attribute_name,
            strategy = self.load,
            mapping = self.mapping,
            table = model.table,
            join_column = fk,
            secondary_table = join_table,
            secondary_join_column = counter_fk,
        )

        model.add_relationship(relationship)

        if self.master:
            InsertEntityEvent(model).add_listener(PersistAssociationEventListener(relationship)) 


from sideral.mapper import Model
from sideral.mapper import properties
from sideral.schema import Table
from sideral.schema import Column
from sideral.schema import ForeignKey
from sideral.errors import ReferenceNotInstantiated
from sideral.event import InsertEntityEvent, PersistAssociationEventListener


from __future__ import annotations
from sideral.utils import abstract


class Event:
    _listeners: list[EventListener]

    def add_listener(self, listener: EventListener) -> None:
        self._listeners.append(listener)

    def post_event(self, *args, **kwargs) -> None:
        for listener in self._listeners:
            listener(*args, **kwargs)

@abstract
class EntityEvent(Event):
    _instances: dict[str, EntityEvent]

    def __new__(cls, model: Model) -> EntityEvent:
        name = model._class.__name__
        if name in cls._instances:
            return cls._instances[name]
        else:
            event = object.__new__(cls)
            event.init(model)
            cls._instances[name] = event
            return event
        
    def init(self, model: Model) -> None:
        self._model = model
        self._listeners = []
    
    def post_event(self, element: object) -> None:
        for listener in self._listeners:
            listener(model = self._model, element = element)


class UpdateEntityEvent(EntityEvent):
    _instances: dict[str, EntityEvent] = {}

class InsertEntityEvent(EntityEvent):
    _instances: dict[str, EntityEvent] = {}

class DeleteEntityEvent(EntityEvent):
    _instances: dict[str, EntityEvent] = {}


from .listeners import EventListener
from sideral.mapper import Model


from __future__ import annotations
from functools import wraps
from .session import transaction
from .session import SessionFactory
from .session import Session



def repository(_class: type):

    def wrapper(cls):

        @wraps(cls, updated = ())
        class Repository(cls):

            def _get_session(self) -> Session:
                return SessionFactory.get()
            
            @transaction
            def select_all(self) -> list[_class]:
                return self._get_session().load(_class)
            
            @transaction
            def select_by_id(self, id: any) -> _class:
                return self._get_session().load(_class, id = id)
            
            @transaction
            def save(self, element: _class) -> _class:
                session = self._get_session()
                session.register_new(element)
                session.flush()
                return session.load(_class, Model(_class).identifier.get_attribute(element))
            
            @transaction
            def save_all(self, elements: list[_class]) -> _class:
                return [self.save(element) for element in elements]
                
            @transaction
            def delete(self, element: _class) -> None:
                session = self._get_session()
                session.register_removed(element)

        return Repository

    
    return wrapper
            


from .mapper import Model
from .event import InsertEntityEvent, PersistAssociationEventListener


from enum import Enum



class Join(Enum):

    INNER = 'inner join'
    LEFT = 'left outer join'
    RIGHT = 'right outer join'


def format(value: any):
    if isinstance(value, str):
        return f"0x{value.encode('utf-8').hex()}" if value else "''"
    
    if value is None or value is ...:
        return 'null'
    
    return value


from __future__ import annotations
from typing import Generator


class Table:

    _instances: dict[str, Table] = {}
    __slots__ = 'name', 'primary_key', '_columns', '_foreign_keys'

    def __new__(cls, name: str) -> Table:
        try:
            return cls._instances[name]
        except KeyError:
            table = object.__new__(cls)
            table.init(name)
            return table

    def __str__(self) -> str:
        return self.name
    
    def __hash__(self) -> int:
        return hash(self.name)

    def init(self, name: str) -> None:
        self.name = name
        self.primary_key = None
        self._columns = {}
        self._foreign_keys = {}

    def add_column(self, column: Column) -> None:
        column.table = self
        self._columns[column.key] = column

    def add_foreign_key(self, foreign_key: ForeignKey) -> None:
        self._foreign_keys[foreign_key.column.key] = foreign_key

    @property
    def columns(self) -> Generator[Column]:
        return self._columns.values()

    @property
    def foreign_keys(self) -> Generator[ForeignKey]:
        return self._foreign_keys.values()

    def get_column(self, name: str) -> Column:
        try:
            return self._columns[name]
        except KeyError:
            return None

    def get_foreign_key(self, name: str) -> ForeignKey:
        try:
            return self._foreign_keys[name]
        except KeyError:
            return None
    
    def get_alias(self, index: int) -> str:
        return f'{self.name}_{index}'

    def build_select(self) -> Select:
        return (
            Select()
            .columns(self.columns)
            .from_(self)
        )


class Column:

    __slots__ = 'name', 'table', 'auto_increment'
    
    def __init__(self, name: str, auto_increment: bool = False) -> None:
        self.name = name
        self.auto_increment = auto_increment
    
    def __str__(self) -> str: 
        return self.full_name

    def __eq__(self, other: any) -> str: 
        return SQLCondition(self, '=', other)

    def __gt__(self, other: any) -> str: 
        return SQLCondition(self, '>', other)

    def __ge__(self, other: any) -> str: 
        return SQLCondition(self, '>=', other)

    def __lt__(self, other: any) -> str: 
        return SQLCondition(self, '<', other)

    def __le__(self, other: any) -> str: 
        return SQLCondition(self, '<=', other)
    
    def __hash__(self) -> int: 
        return hash(self.full_name)

    @property
    def key(self) -> str:
        return self.name

    @property
    def full_name(self) -> str:
        return f'{self.table.name}.{self.name}'
    
    @property
    def alias(self) -> str:
        return f'{self.table.name}_{self.name}'
    
    def get_alias(self, query: Query) -> str:
        table_alias = ...
        try:
            table_alias = query._tables[self.table][query._index]
        except IndexError:
            table_alias = query._tables[self.table][-1]            
        except KeyError:
            return self.alias

        return f'{table_alias}.{self.name}'   

    def build_select(self) -> Select:
        return (
            Select()
            .columns(self)
            .from_(self.table)
        )
    

class ForeignKey:

    __slots__ = 'column'
    
    def __init__(self, column: Column) -> None:
        self.column = column


class PrimaryKey(ForeignKey):

    def __init__(self, column: Column) -> None:
        super().__init__(column)
    

class ForeignKey(ForeignKey):

    __slots__ = 'reference'

    def __init__(self, column: Column, reference: Column) -> None:
        super().__init__(column)
        self.reference = reference


from .query import Query, Select, format, SQLCondition


from __future__ import annotations

    


class SQLCondition:

    def __init__(self, column: Column, operator: str, value: any) -> None:
        self._column = column
        self._operator = operator
        self._value = value
    
    def __str__(self) -> str:
        return f'{self._column} {self._operator} {format(self._value)}'
    
    def to_string(self, query: Query = None, on: bool = False) -> str:
        if on:
            column1 = self._column.get_alias(query)
            query._index = 1
            column2 = self._value.get_alias(query)
            return f'{column1} {self._operator} {column2}'
        return f'{self._column.get_alias(query)} {self._operator} {format(self._value)}'



from sideral.schema import Column
from .builder import Query
from .utils import format


from __future__ import annotations
from .annotations import annotation
from .utils import _hash
from sideral.strategy import  load
from collections.abc import Callable
from typing import TypeVar, ParamSpec


_params = ParamSpec('_params')
_return = TypeVar('_return')


class column(annotation):

    def __init__(self, function: Callable[_params, _return] = None, name: str = None, load: load = load.EAGER) -> None:
        annotation.__init__(self, function, _name = name, load = load)
    
    def __set_name__(self, owner: type, name: str) -> None:
        super().__set_name__(owner, name)
        try:
            owner.__dict__['__columns__'].append(name)
        except KeyError:
            owner.__columns__ = [name]
        
    @property
    def name(self) -> str:
        return self._name or self.get.__name__
    
    def map(self) -> None:
        model = Model(self.owner)
        column = Column(name = self.name, auto_increment = False)
        model.table.add_column(column)
        model.add_column(properties.Column(attribute_name = self.attribute_name, strategy = self.load, column = column))


class id(column):

    def __init__(self, function: Callable[_params, _return] = None, name: str = None, auto_increment: bool = False) -> None:
        annotation.__init__(self, function, _name = name, auto_increment = auto_increment)

    def __set_name__(self, owner: type, name: str) -> None:
        super().__set_name__(owner, name)
        setattr(owner, '__hash__', _hash(name))

    def map(self) -> None:
        model = Model(self.owner)
        column = Column(name = self.name, auto_increment = self.auto_increment)
        model.table.add_column(column = column)
        model.table.primary_key = PrimaryKey(column = column)
        model.identifier = properties.Column(attribute_name = self.attribute_name, strategy = load.EAGER, column = column)


from sideral.mapper import Model
from sideral.mapper import properties
from sideral.schema import Column
from sideral.schema import PrimaryKey


from __future__ import annotations
from sideral.errors import UnloadedIterator
from typing import Generic, TypeVar

def decorator(callable):
    return callable




class _Unloaded: 
    
    _instance: _Unloaded

    def __new__(cls) -> _Unloaded:
        try:
            return cls._instance
        except AttributeError:
            cls._instance = object.__new__(cls)
            return cls._instance

    def __iter__(self) -> None: 
        raise UnloadedIterator
    
    def __repr__(self) -> str:
        return f'<UNLOADED>'



T = TypeVar('T')
class ordered_set(Generic[T]):
    
    def __init__(self, *items) -> None:
        self._items = {}
        self._hashs = []

        for item in items:
            h = hash(item)
            self._items[h] = item
            self._hashs.append(h)
    
    def __str__(self) -> str:
        return str([self._items[h] for h in self._hashs])
    
    def __iter__(self) -> ordered_set:
        self._index = -1
        return self
    
    def __next__(self) -> T:
        self._index += 1

        if self._index > len(self._hashs) - 1:
            raise StopIteration
        
        return self._items[self._hashs[self._index]]
    
    def __len__(self) -> int:
        return len(self._hashs)
    
    def __contains__(self, item: any) -> bool:
        return (hash(item) in self._items)
    
    def __getitem__(self, index: int) -> any:
        return self._hashs[index]
    
    def add(self, item: any) -> None:
        h = hash(item)

        if h not in self._items:
            self._hashs.append(h)

        self._items[h] = item
    
    def remove(self, item: any) -> None:
        h = hash(item)
        self._hashs.remove(h)
        del self._items[h]
            


    

abstract = decorator
metaclass = decorator
interface = decorator
override = decorator
UNLOADED = _Unloaded()


from __future__ import annotations
from sideral.utils import interface
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .builder import Select

@interface
class Selectable:
    
    def build_select(self) -> Select: ...


@interface
class Modifiable:
    
    def save(self) -> Select: ...
